/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensio.ns, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * This file contains main program which calls the file-system browser
 * (main event loop)
 *****************************************************************************/

#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include "fm.h"

static char SccsID[] = "@(#)fm.c	1.4\t1/27/93";            /* sccs keyword */

dirWndHead *Head = NULL;
dirWindow *toRemove = (dirWindow *) NULL;

main(argc,argv)
     int argc;
     char *argv[];
{
  char *getcwd();
  int forcedQuit();
  Display *display;
  Widget top;
  XEvent event;

  int c;
  int err = 0;
  int x = DEF_XPOS;
  int y = DEF_YPOS;
  extern int optind, opterr;
  extern char *optarg;
  int eshErrorHandler();
  int updateDirectories();

  char *dir = NULL;
  static char *usage = "fm [-x <num> -y <num> -f <file>";
  opterr = 0;
  
  /* parse the command line arguments */
  while ((c = getopt(argc, argv, "f:x:y:")) != EOF) {
    switch (c) {
    case 'f':                       /* starting directory */
      dir = optarg;
      break;
    case 'x':                       /* x-position of upper left corner */
      x = atoi(optarg);
      break;
    case 'y':                       /* y-position of upper left corner */
      y = atoi(optarg);
      break;
    case '?':                       /* default reaction on arguments
				       not defined */
      err = 1;
    }
    if (err) {
      fatal("FileSystemBrowser", usage, NULL);
    }
  }

  top = XtInitialize(NULL, "FM", (XrmOptionDescRec *) NULL, 0, &argc, argv);

  display = XtDisplay(top);
  XSynchronize(display, True); 
  XSetErrorHandler(eshErrorHandler);
  
  signal(SIGALRM, updateDirectories); 
  
  initFM(top);

  if (! dir) 
    dir = getcwd(NULL, 100);

  initDir(top, x, y, dir);
  
  alarm(0);

  while (True) {
    if (toRemove) {
      removeDirWindow(toRemove);
      toRemove = (dirWindow *) NULL;
    }

    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
}

/*ARGSUSED*/
int eshErrorHandler(dsp, event)
     Display *dsp;
     XErrorEvent *event;
{
  kill(getpid(), SIGQUIT);
}

/* quitFm: ********************************************************************
 * quit (what = FMQUIT) or abort (what = FMABORT) FileTreeBrowser
 *****************************************************************************/

quitFm(wnd, what)
     dirWindow *wnd;
     int what;
{
  
  if (what == FMQUIT)
    commitChanges(wnd, True, False);

  XCloseDisplay(XtDisplay(Head->root));

  exit(0);
}

