/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <X11/Intrinsic.h>

#include <exBasics.h>
#include <exLayout.h>

#include "exEdit.h"
#include "display.h"
#include <exEntry.h>

#include <stdio.h>

static char SccsID[] = "@(#)Einterface.c	1.3\t11/10/92";            /* sccs keyword */

BufInfo *bufinfoptr();
static char msg[200];

char *ETtext();
Boolean ETemptybuf();
Boolean phys2screen();
char l[MAXLINE];

/****************************************************************************
 *               editor functions (binding to command table possible)
 ****************************************************************************/

/* right: *******************************************************************
 * move cursor right, advance to next line if end of line is reached
 ****************************************************************************/

right(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos < ETlen(bf, bf->curline)) {	/* advance in line */
    bf->curpos++;
    if (! phys2screen(bf, bf->curline, bf->curpos, 
		      &bf->currow, &bf->curcol, True))
      ESscroll(bf, 1, DIR_UP);
  } else if (bf->curline < ETnumlines(bf) - 1) 	/* advance to next line */
    dwnlin(editor);
}

/* left: *********************************************************************
 * move cursor left, go to previous line if beginning of line is reached
 *****************************************************************************/

left(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos > 0) {
    bf->curpos--;
    setcol(bf);
  } else if (bf->currow > 0) {
    uplin(editor);
    endlin(editor);
  }
}

/* beglin: *******************************************************************
 * go to beginning of line 
 *****************************************************************************/

beglin(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  bf->curpos = 0;
  setcol(bf);
}

/* endlin: *******************************************************************
 * go to end of line; scroll if necessary
 *****************************************************************************/

endlin(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int curline = bf->curline, currow = bf->currow, curcol = bf->curcol;

  bf->curpos = ETlen(bf, bf->curline);
  setcol(bf);

  if (bf->curline < curline) {
    bf->curline = curline;
    bf->currow = currow;
    bf->curcol = curcol;
    if (ESscroll(bf, topLine(bf, bf->curline) - bf->topline, DIR_UP) < 0)
      setpos(bf); 
  }
}

/* dwnlin: *******************************************************************
 * go one line down; position cursor at beginning of line; scroll if necessary
 *****************************************************************************/

dwnlin(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int screen_lines = screenLines(bf, bf->curline) - screenLine(bf);
  int dummy;

  if (bf->curline >= ETnumlines(bf) - 1)		/* bottom of buffer */
    return;

  if (bf->currow < ESheight(bf) - screen_lines) {
    bf->currow += screen_lines;
    bf->curline++;
  } else if (! phys2screen(bf, ETnumlines(bf), 0, &dummy, &dummy, False)) {
    ESscroll(bf, 1, DIR_UP);
    bf->curline = MIN(ETnumlines(bf) - 1, bf->curline + 1);
  }
  
  bf->curcol = 0;
  bf->curpos = 0;
  bf->lmarg = 0;

  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
}

/* uplin: ********************************************************************
 * go one line up; position cursor at end of line; scroll if necessary
 ****************************************************************************/

uplin(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor); 
  int screen_lines = 0; 

  if(bf->curline > 0)
    screen_lines = screenLines(bf, bf->curline - 1);

  if (bf->currow > 0) {
    bf->currow -= screen_lines;
    bf->curline--;
  } else if (bf->topline > 0) {
    ESscroll(bf, 1, DIR_DOWN);
    bf->curline = MAX(0, bf->curline - 1);
  }

  endlin(editor);
}

/* dwncol: ********************************************************************
 * go down one line, preserve column
 ***************************************************************************/

dwncol(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int screen_lines = screenLines(bf, bf->curline) - screenLine(bf);
  int dummy;

  if (bf->curline >= ETnumlines(bf) - 1)		/* bottom of buffer */
    return;

  preserveCol(bf); 
  if (bf->currow < ESheight(bf) - screen_lines) {
    bf->currow += screen_lines;
    bf->curline++;
    if (! phys2screen(bf, bf->curline, bf->curpos, &dummy, &dummy, False))
      ESscroll(bf, 1, DIR_UP);
  } else if (! phys2screen(bf, ETnumlines(bf), 0, &dummy, &dummy, False)) {
    ESscroll(bf, 1, DIR_UP);
    bf->curline = MIN(ETnumlines(bf) - 1, bf->curline + 1);
  }

  setPreservedCol(bf);
  setcol(bf);
  setpos(bf);
}

/* upcol: ********************************************************************
 * go up in column 
 ****************************************************************************/

upcol(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int screen_lines; 

  if (bf->curline <= 0)		/* already in first line */
    return;

  screen_lines = screenLines(bf, bf->curline - 1);

  preserveCol(bf); 

  if (bf->currow >= screen_lines) {
    bf->currow -= screen_lines;
    bf->curline--;
  } else if (bf->topline > 0) {
    ESscroll(bf, 1, DIR_DOWN);
    bf->curline = MAX(0, bf->curline - 1);
  }

  setPreservedCol(bf);
  setpos(bf);
}

/* upScroll: ****************************************************************
 * scroll one line up
 ****************************************************************************/

upScroll(editor)
     Editor editor;
{
  scrollLines(bufinfoptr(editor), 1, DIR_UP);
}

/* dwnScroll: *****************************************************************
 * scroll down one line 
 *****************************************************************************/

dwnScroll(editor)
     Editor editor;
{
  scrollLines(bufinfoptr(editor), 1, DIR_DOWN);
}

/* dwnpage: ******************************************************************
 * scroll down a screen 
 ****************************************************************************/

dwnpage(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  scrollLines(bf, bf->lines, DIR_DOWN);
}  

/* uppage: *******************************************************************
 * scroll up a screen 
 ****************************************************************************/

uppage(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  scrollLines(bf, bf->lines, DIR_UP);
}

/* uphpage: *******************************************************************
 * scroll up half a screen 
 *****************************************************************************/

uphpage(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  scrollLines(bf, bf->lines / 2, DIR_UP);
}

/* dwnhpage: ******************************************************************
 * scroll down half a screen 
 *****************************************************************************/

dwnhpage(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  scrollLines(bf, bf->lines / 2, DIR_DOWN);
}

/* wordr: *********************************************************************
 * make cursor skip one word to the right, advance to next line if end of line
 * is reached
 *****************************************************************************/

wordr(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos >= ETlen(bf, bf->curline))
    dwnlin(editor);
  else
    ESwords(bf, DIR_RIGHT);
}

/* wordl: *********************************************************************
 * make cursor skip one word to the left, advance to previous line if end of 
 * line is reached 
 *****************************************************************************/

wordl(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos == 0) 
    uplin(editor);
  else
    ESwords(bf, DIR_LEFT);
}

/* botscr: ********************************************************************
 * position cursor to last line of the window
 *****************************************************************************/

botscr(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  register numlines = ETnumlines(bf);

  preserveCol(bf); 
  bf->currow = MIN(bf->lines - 1, (numlines - bf->topline) - 1);
  setPreservedCol(bf);
  setpos(bf);
}

/* topscr: ********************************************************************
 * position cursor to first line of the window
 *****************************************************************************/

topscr(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  preserveCol(bf); 
  bf->currow = 0;
  setPreservedCol(bf);
  setpos(bf);
}

/* midscr: ********************************************************************
 * position cursor to line in the middle of the window
 *****************************************************************************/

midscr(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  register numlines = ETnumlines(bf);

  preserveCol(bf); 
  bf->currow = MIN((bf->lines / 2) - 1, (numlines - bf->topline) - 1);
  setPreservedCol(bf);
  setpos(bf);
}

/* topBuffer: ****************************************************************
 * position the cursor at the beginning of the first line; if necessary 
 * the screen is scrolled, so that the first text line appears at the 
 * top of the window
 ****************************************************************************/

topBuffer(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->topline == 0 && bf->currow == 0 && bf->curcol == 0)
    return;
  
  ESshowcur(bf, False, bf->curcol, bf->currow);
  csrChrForget(bf);
  showline(bf, 0, 0);
  bf->topline = bf->curline = bf->curpos = bf->currow = bf->curcol = 0;
  ESshowcur(bf, True, 0, 0);

}

/* botBuffer: *****************************************************************
 * position the cursor in the last line; if necessary the screen is scrolled, 
 * so that the last text line appears in the middle of the window
 *****************************************************************************/
 
botBuffer(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int line = ETnumlines(bf) - 1;

  ESshowcur(bf, False, bf->curcol, bf->currow);
  csrChrForget(bf);
  showline(bf, line, strlen(ETtext(bf, line)));
  setcol(bf);
  ESshowcur(bf, True, bf->curcol, bf->currow);
}

/* lineToTop: *****************************************************************
 * scroll text in window, so that the current line is the first line in window
 *****************************************************************************/

lineToTop(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  ESscroll(bf, bf->curline - bf->topline, DIR_UP);
  bf->curline = bf->topline;
  setcol(bf);
}

/* delnxt: *******************************************************************
 * delete character under cursor position; append next line if cursor is at the
 * end of the line
 *****************************************************************************/

delnxt(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos == ETlen(bf, bf->curline))
    joinlines(bf);
  else 
    delchar(bf);
}

/* delprev: *******************************************************************
 * delete character before cursor position; append current line to previous 
 * line if cursor is at the beginning of the line
 *****************************************************************************/

delprev(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->curpos > 0) {
    bf->curpos--;
    setcol(bf);
    delchar(bf);
    bf->oldcol = bf->curcol;
  }
  else if (bf->curline >= 1) {

    if (bf->curline == bf->topline)
      ESscroll(bf, 1, DIR_DOWN);

    bf->curline--;
    bf->curpos = ETlen(bf, bf->curline);
    setcol(bf);
    joinlines(bf);
  }
}

/* fdelword: ******************************************************************
 * delete next word: 
 *****************************************************************************/

fdelword(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  register char *t = ETtext(bf, bf->curline);
  register cnt = 0;

  if (t[bf->curpos] == ' ') {
    l[cnt++] = t[bf->curpos];
    delnxt(editor);
  }

  if (t[bf->curpos] == EOS) { 		/* cursor at end of line */
    delnxt(editor);			/* basically: do a joinlines() */
    delnxt(editor);
    return;
  }
  
  while (t[bf->curpos] != ' ' && t[bf->curpos] != EOS) {
    l[cnt++] = t[bf->curpos];
    delnxt(editor);			/* delete chars */
  }

  l[cnt] = '\0';
  setKillBuffer(bf, strCopy(l));
}

/* bdelword: ******************************************************************
 * delete previous word
 *****************************************************************************/

bdelword(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int line; 
  register char **tb = bf->bufPtr->text;
  register cnt = sizeof(l) - 1;

  l[cnt--] = '\0';
  line = bf->curline;

  while (bf->curpos == 0) {		/* delete empty lines */
    if (line == 0)
      return;
    else {
      delprev(editor);
      l[cnt--] = tb[line][bf->curpos];
    }
  }

  while (tb[line][bf->curpos - 1] == ' ' || tb[line][bf->curpos - 1] == '\t') {
    delprev(editor);
    l[cnt--] = tb[line][bf->curpos - 1];
    while (bf->curpos == 0) {				/* start of line */
      if (bf->topline == 0) {				/* top of buffer */
	setKillBuffer(bf, strCopy(&l[cnt + 1]));
	return;
      } else {						/* join lines */
	delprev(editor);
	l[cnt--] = tb[line][bf->curpos - 1];
	line = bf->curline;				/* recalculate */
      }
    }
  }  

  while (tb[line][bf->curpos - 1] != ' ' && tb[line][bf->curpos - 1] != '\t') {
    delprev(editor);
    l[cnt--] = tb[line][bf->curpos - 1];
    if (bf->curpos == 0) {				/* start of line */
      break;
    }
  }
  setKillBuffer(bf, strCopy(&l[cnt + 1]));
}

/* krest: *********************************************************************
 * delete the rest of the line including the current position; append next 
 * line if cursor is at the end of the line 
 *****************************************************************************/

krest(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  int line, sl, row;
  RECT rect;

  if (ETemptybuf(bf))
    return;

  line = bf->curline;
  row = screenLine(bf);

  strcpy(l, ETtext(bf, line));
  if (l[bf->curpos] != EOS) {             	/* delete text */
    setKillBuffer(bf, strCopy(&bf->bufPtr->text[line][bf->curpos]));
    sl = screenLines(bf, bf->curline);
    l[bf->curpos] = EOS;
    ETreplace(bf, line, l);
    csrChrForget(bf);

    if (row == sl - 1)  			/* last row of displ. line */
      PSclear(bf, bf->currow, bf->curcol);	/* clear rest of cur. row */
    else {					/* clear more than one row */
      PSclear(bf, bf->currow, bf->curcol);	/* clear first line */
      if (bf->currow + sl - (row + 1) < (bf->lines - 1)) { /* rest existing
							      and visible */
	RASG(&rect, bf->currow + sl - (row + 1), 0, bf->lines - 1, bf->cols); 
	PSslide(bf, &rect, DIR_UP, sl-(row + 1)); /* move next lines up */
	PSclear(bf, bf->currow, 0);		  /* clear new first line */
	showRangeOfRows(bf, bf->lines - sl - (row + 1), bf->lines - 1);
      }
      else {  /* text to be deleted reaches bottom of screen */
	RASG(&rect, bf->currow + 1, 0, bf->lines - 1, bf->cols);
	PSfill(bf, &rect);
	showRangeOfRows(bf, bf->currow + 1, bf->lines - 1);
      }
    }
  } else if (ETnumlines(bf) > 1)              	/* combine lines */
    joinlines(bf);
}

/* newline: *******************************************************************
 * insert a newline-character at the current position thereby splitting 
 * the line 
 *****************************************************************************/

newline(editor)
     Editor editor;
{
  splitline(bufinfoptr(editor));
}

/* openln: ********************************************************************
 * open a new line above the current line 
 *****************************************************************************/

openln(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor); 

  csrChrForget(bf);
  ESshowcur(bf, False, bf->curcol, bf->currow);

  insline(bf);

  ESsetcur(bf);
}

/* openlnd: *******************************************************************
 * open a line below the current line
 *****************************************************************************/

openlnd(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor); 

  csrChrForget(bf);
  ESshowcur(bf, False, bf->curcol, bf->currow);

  bf->curline++;
  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, False);

  insline(bf);

  ESsetcur(bf);
}

/* fsearch: *******************************************************************
 * search forward: ask user for search string
 *****************************************************************************/

fsearch(editor)
     Editor editor;
{
  register char *t;
  register BufInfo *bf = bufinfoptr(editor);
  
  t = exGetEntry(bf->w, -1, -1, "Search: ", bf->sbuff, 20);
  if (t) {
    strcpy(bf->sbuff, t);
    free(t);
  } else
    bf->sbuff[0] = EOS;

  ESsearch(bf, DIR_DOWN);
}

/* bwsearch: ******************************************************************
 * search backward: ask user for search string
 *****************************************************************************/

bwsearch(editor)
     Editor editor;
{
  register char *t;
  register BufInfo *bf = bufinfoptr(editor);

  t = exGetEntry(bf->w, -1, -1, "Search backward: ", bf->sbuff, 20);
  if (t) {
    strcpy(bf->sbuff, t);
    free(t);
  } else
    bf->sbuff[0] = EOS;

  ESsearch(bf, DIR_UP);
}

/* ffsearch: *****************************************************************
 * search forward: use last search string if any
 *****************************************************************************/

ffsearch(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (! bf->sbuff[0])		 	/* no search string specified yet*/
    fsearch(editor);
  else					/* use last search string */
    ESsearch(bf, DIR_DOWN);
}

/* bbsearch: ******************************************************************
 * search backward: use last search string if any
 *****************************************************************************/

bbsearch(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (! bf->sbuff[0])		 	/* no search string specified yet*/
    bwsearch(editor);
  else					/* use last search string */
    ESsearch(bf, DIR_UP);
}

/* newscr: ********************************************************************
 * update screen
 *****************************************************************************/

newscr(editor)
     Editor editor;
{
  ESshowall(bufinfoptr(editor));
}

/* memory: ********************************************************************
 * display statistics: cursor position relative to the screen and the buffer,
 * the maximal number of lines of the current buffer and the identifier of this
 * buffer.
 *****************************************************************************/

memory(editor)
     Editor editor;
{
  status(bufinfoptr(editor));
}

/* findFile: ******************************************************************
 * read a file into the editor buffer
 *****************************************************************************/

findFile(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  char *filename;

  filename = exGetFilename(bf->w, -1, -1, "Read file: ", NULL, MAXFNAME);

  if (! filename || ! *filename)
    return(-1);

  if (strlen(filename) > MAXFNAME) {
    ESmsg(bf, "File name too long !", True);
    free(filename);
    return(-1);
  }

  bf->filename = filename;

  readfile(bf);
  return(1);
}

/* insertFile: ****************************************************************
 * insert the user specified file into the buffer at the current cursor 
 * position; scroll until last line of inserted file is visible on the screen
 *****************************************************************************/

insertFile(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  char *filename;

  filename = exGetFilename(bf->w, -1, -1, "Insert file: ", NULL, MAXFNAME);

  if (! filename || ! *filename)
    return(-1);

  if (strlen(filename) > MAXFNAME) {
    ESmsg(bf, "File name too long !", True);
    free(filename);
    return(-1);
  }

  insertfile(bf, filename);
  
  return(1);
}

/* killBuffer: ****************************************************************
 * clear the current buffer after user confirmation 
 *****************************************************************************/

killBuffer(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  killbuf(bf, False);
  csrChrForget(bf);
}

/* clearBuffer: ***************************************************************
 * clear the current buffer without asking the user 
 *****************************************************************************/

clearBuffer(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  killbuf(bf, True);
  csrChrForget(bf);
}

/* saveFile: ****************************************************************
 * store the buffer contents into corresponding file if any; otherwise the
 * user is asked for a file name
 ****************************************************************************/

saveFile(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (!bf->filename || !*bf->filename)	/* no filename provided */
    writeFile(editor);			/* ask for a file name */
  else {
    writefile(bf);			/* save file */
    sprintf(msg, "Buffer saved -- %s -- !", bf->filename);
    ESmsg(bf, msg, True);
  }
}

/* writeFile: ****************************************************************
 * ask the user for a file name in which to store the buffer contents
 *****************************************************************************/

writeFile(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  char *file;

  file = exGetFilename(bf->w, -1, -1, "Write to file: ", NULL, MAXFNAME);

  if (! file || ! *file)
    return;

  free(bf->filename);
  bf->filename = file;
  writefile(bf);

  sprintf(msg, "Buffer written to file -- %s -- !", bf->filename);
  ESmsg(bf, msg, True);
}

/* help: *******************************************************************
 * display the help editor if any, else notify the user that no help is 
 * available
 ***************************************************************************/

help(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  int x, y;

  if (bf->hlp) {
    getMouse_xy(XtDisplay(bf->w), &x, &y);
    topscr(bf->hlp);
    stxtShow(bf->hlp, x, y);
  } else
    ESmsg(bf, "No online help available !", True);
}

/* resetViewMode: *************************************************************
 * turn view mode off
 *****************************************************************************/

resetViewMode(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  bf->flags &= ~ED_VIEWMODE;
}

/* setViewMode: **************************************************************
 * turn view mode on 
 ****************************************************************************/

setViewMode(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  bf->flags |= ED_VIEWMODE;
}

/* editAbort: *****************************************************************
 * abort the editor 
 *****************************************************************************/

editAbort(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  sprintf(msg, "ABORT editor -- %s -- ?", bufinfoptr(editor)->name);
  if (exNotifyUser(bf->w, -1, -1, msg) == ANSWERYES)
    exit(1);
}

/* editExit: ******************************************************************
 * exit gracefully 
 *****************************************************************************/

editExit(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (/*! ETemptybuf(bf) &&*/ ETchanged(bf)) {  /* buffer modified */
    sprintf(msg, "Save modified buffer\nbefore exiting editor -- %s --?", 
	    bf->name);
      switch (exNotify2User(bf->w, -1, -1, msg)) {
      case ANSWERYES:
	saveFile(editor);
	break;
      case ANSWERCANCEL:
	return;
      default:
	break;
      }
  }
  else {                                  /* no changes after last save */
    sprintf(msg, "Exit editor -- %s -- ?", bf->name);    
    switch (exNotifyUser(bf->w, -1, -1, msg)) {
    case ANSWERNO:
      return;
    default:
      break;
    }
  }

  exit(1);
}
 
/* goLine: ********************************************************************
 * goto the specified physical line 
 *****************************************************************************/

goLine(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  char *lineNo;
  int line;

  lineNo = exGetEntry(bf->w, -1, -1, "what line: ", NULL, 5);

  if (lineNo && *lineNo) {
    line = atoi(lineNo);
    free(lineNo);
  } else 
    line = -1;

  showline(bf, line, 0);
}

/* whatLine: ******************************************************************
 * tell the current line
 *****************************************************************************/

whatLine(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  
  sprintf(msg, "Current line is -- %d,%d (%d,%d)-- !", 
	  bf->curline, bf->curpos, bf->currow, bf->curcol);
  ESmsg(bf, msg, True);
}

/* noop: *********************************************************************
 * just do nothing
 *****************************************************************************/

noop(editor)
     Editor editor;
{
  ESmsg(bufinfoptr(editor), "Illegal command!", False);
}

/* setFocus: ******************************************************************
 * set the input focus to the given editor's editor window
 *****************************************************************************/

setFocus(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);

  XWarpPointer(XtDisplay(bf->edt), None, XtWindow(bf->edt), 0, 0, 0, 0, 
	       exFontWidth(bf->font) / 2, LINEHEIGHT(bf->font) / 2);
}
  
/* putCursor: ****************************************************************
 * calls ESsetcur to reposition the cursor 
 *****************************************************************************/

putCursor(editor)
     Editor editor;
{
  ESsetcur(bufinfoptr(editor));
}

/* copy: *********************************************************************
 * copy the text between cursor and dot to the kill buffer
 ****************************************************************************/

copy(editor)
     Editor editor;
{
  exCopy(bufinfoptr(editor));
}

/* cut: **********************************************************************
 * copy the text between mark and dot to the kill buffer and delete the 
 * marked text in th text editor buffer 
 ****************************************************************************/

cut(editor)
     Editor editor;
{
  exCut(bufinfoptr(editor));
}

/* dot: *********************************************************************
 * set the dot to the current position; the region between this position and
 * a later current position may be cut and copied 
 ****************************************************************************/

dot(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  
  setDot(bf, bf->curcol, bf->currow);
}

/* exMark: *******************************************************************
 * if the dot is defined set the mark to the current position; thus, a region
 * is defined
 ****************************************************************************/

exMark(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  
  if (bf->region.d_line < 0) {
    ESmsg(bf, "Cannot set mark: define a dot first !", True);
    Bell();
    return;
  }

  setMark(bf, bf->curcol, bf->currow);		/* set mark */
  ESdisplayRegion(bf, True);			/* display region */
}

/* yank: *********************************************************************
 * yank the contents of the kill buffer to the current position
 ****************************************************************************/

yank(editor)
     Editor editor;
{
  exYank(bufinfoptr(editor));
}

/* exchangeDotMark: *********************************************************
 * set mark to the current position; then exchange dot and mark and position
 * cursor to previous dot;
 ****************************************************************************/

exchangeDotMark(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  register struct Region *r = &bf->region;

  if (bf->region.d_line < 0) {
    ESmsg(bf, "No dot set !", True);
    Bell();
    return;
  }

  if (bf->region.m_line < 0) {
    ESmsg(bf, "No mark set !", True);
    Bell();
    return;
  }

  dot2mark(bf);

  if (bf->curpos == r->d_pos)			/* cur. pos is dot-pos */
    showline(bf, r->m_line, r->m_pos);  	/* show line with mark */
  else
    showline(bf, r->d_line, r->d_pos);  	/* show line with dot */

  ESsetcur(bf);
}

/* resetRegion: **************************************************************
 * if a region is defined reset it
 ****************************************************************************/

resetRegion(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  ESdisplayRegion(bf, False);
  reset_region(bf);
}

/****************************************************************************
 *               editor utility functions
 *
 * (binding to command table not possible because additional parameters are
 * required or function provides result values)
 *
 * - get/write text to/from editor buffer
 * - get/set internal editor values and create editor
 ****************************************************************************/

/*****************************************************************************
 *               get/write text to/from editor buffer
 *****************************************************************************/

/* emptyBuffer: **************************************************************
 * returns TRUE if text buffer is empty
 *****************************************************************************/

Boolean emptyBuffer(editor)
     Editor editor;
{
  return(ETemptybuf(bufinfoptr(editor)));
}
 
/* getText: ******************************************************************
 * returns the text of the given line
 *****************************************************************************/

char *getText(editor, line)
     Editor editor;
     int line;
{
  char *ETtext();

  register BufInfo *bf = bufinfoptr(editor);

  if (line < 0 || line > ETnumlines(bf))
    return((char *) 0);

  return(ETtext(bf, line));
}

/* getCurrentLine: ***********************************************************
 * get the text of the line to which cursor points
 ****************************************************************************/

char *getCurrentLine(editor)
     Editor editor;
{
  register BufInfo *bf = bufinfoptr(editor);
  
  return(bf->bufPtr->text[bf->curline]);
}

/* getCurrentWord: ************************************************************
 * return the current word (the word under (or just before) the cursor)
 *****************************************************************************/

char *getCurrentWord(editor)
     Editor editor;
{
  Boolean locateCurrentWord();
  char *res;
  char *lpos, *rpos;

  if ( !locateCurrentWord(editor, &lpos, &rpos))
    return(NULL);

  res = (char *) malloc(rpos - lpos + 2);
  bcopy(lpos, res, rpos - lpos + 1);
  res[rpos - lpos + 1] = EOS;

  return(res);
}

/* getBufferContents: *********************************************************
 * return the buffer contents; append to each line a newline.
 *****************************************************************************/

char *getBufferContents(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  static char CR[] = {'\n', EOS};
  register char *txt;
  register len, i;
  int numlines = ETnumlines(bf);
  
  for (i = 0, len = 0; i < numlines; i++)
    len += ETlen(bf, i);
  
  txt = (char *) malloc(len + 1 + numlines);
  *txt = EOS;
  
  for (i = 0; i < numlines; i++) {
    strcat(txt, ETtext(bf, i));
    strcat(txt, CR);
  }
  
  return(txt);
}

/* getBufferContents2: ********************************************************
 * return the buffer contents; append to each line a newline besides to last 
 * line.
 *****************************************************************************/

char *getBufferContents2(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  static char CR[] = {'\n', EOS};
  register char *txt;
  register len, i;
  int numlines = ETnumlines(bf);
  
  for (i = 0, len = 0; i < numlines; i++)
    len += ETlen(bf, i);
  
  txt = (char *) malloc(len + 1 + numlines);
  *txt = EOS;
  
  for (i = 0; i < numlines - 1; i++) {
    strcat(txt, ETtext(bf, i));
    strcat(txt, CR);
  }
  strcat(txt, ETtext(bf, numlines - 1));
  
  return(txt);
}

/* getRange: *****************************************************************
 * return buffer area from line1 to line2; append a special character to each 
 * line
 ****************************************************************************/

char *getRange(editor, line1, line2, c)
     Editor editor;
     int line1, line2;
     char c;
{
  static char CR[2];
  register BufInfo *bf = bufinfoptr(editor);
  register char *txt;
  register len, i;

  CR[0] = c;
  CR[1] = EOS;
  
  if (line1 > line2)
    return ((char *) 0);

  for (i = line1, len = 0; i <= line2; i++)
    len += ETlen(bf, i);
  
  txt = Malloc(len + 1 + (line2 - line1));
  
  for (i = line1; i <= line2; i++) {
    strcat(txt, bf->bufPtr->text[i]);

    if (i < line2)
      strcat(txt, CR);
  }
  
  return(txt);
}

/* editorInsertChar: **********************************************************
 * insert the character at the editor's current cursor location 
 *****************************************************************************/

editorInsertChar(editor, ch)
     Editor editor;
     char ch;
{
  register BufInfo *bf = bufinfoptr(editor);

  inschar(bf, ch);

  ESsetcur(bf);
}

/* editorInsert: **************************************************************
 * insert the input string at the current position into the editor buffer and
 * position the cursor at the end of the inserted text
 *****************************************************************************/

editorInsert(editor, txtBuffer)
     Editor editor;
     char *txtBuffer;
{
  register BufInfo *bf = bufinfoptr(editor);

  editorInsertAt(editor, txtBuffer, bf->curline, bf->curpos);
}

/* editorInsertAt: ************************************************************
 * insert the string beginning at 'line' and 'pos' into the editor buffer and
 * position the cursor at the end of the inserted text
 *****************************************************************************/

editorInsertAt(editor, txtBuffer, line, pos)
     Editor editor;
     char *txtBuffer;
     int line, pos;
{
  int oldtop, tline = line;
  register BufInfo *bf = bufinfoptr(editor);

  ESshowcur(bf, False, bf->curcol, bf->currow);

  oldtop = bf->topline;
  (void)insertText(bf, txtBuffer, &tline, pos);

  if (oldtop == bf->topline)                 /* inserted text fits on screen */
    showRangeOfLines(bf, line, bf->curline/*bf->lines - 1*/);
  else {    	/* inserted text does not fit completely on current screen */
    csrChrForget(bf);
    showRangeOfLines(bf, bf->topline, bf->curline);
  }

  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);

  ESsetcur(bf);
}

/* editorAppend: ************************************************************
 * append the string to the end of the editor buffer; position cursor at the
 * end of the appended string
 ****************************************************************************/

editorAppend(editor, txtBuffer)
     Editor editor;
     char *txtBuffer;
{
  int line1, line2, row2, col, pos;
  register BufInfo *bf = bufinfoptr(editor);

  line1 = ETnumlines(bf) - 1;		/* last line in text-buffer */
  appendText(editor, txtBuffer);	/* display not updated */
  ETunchanged(bf);
  line2 = ETnumlines(bf) - 1;		/* last line in buffer after append */

  ESshowcur(bf, False, bf->curcol, bf->currow);
  csrChrForget(bf);

  pos = strlen((char *) EMtext(bf, line2, False));
  phys2screen(bf, line2, pos, &row2, &col, False); 

  if ((row2 < bf->lines) && (col < bf->cols))
                                             /* appended text fits on screen */
    showRangeOfLines(bf, line1, line2);
  else {				    /* scroll appropriately */
    if (! showline(bf, line2, pos))
      showRangeOfLines(bf, line1, line2);
  }
  
  bf->curline = line2;
  bf->curpos = pos;

  phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);

  ESsetcur(bf);
}

/* editorFillBuffer: *********************************************************
 * fill the buffer with the given text; position cursor at the beginning or 
 * end of the buffer
 *****************************************************************************/

editorFillBuffer(editor, txt, end)
     Editor editor;
     char *txt;
     Boolean end;                      /* if True, cursor is positioned at the
					  end of the buffer */
{
  register BufInfo *bf = bufinfoptr(editor);

  killbuf(bf, True);

  if (end)
    editorAppend(editor, txt);
  else {
    appendText(editor, txt);	/* display not updated */
    ETunchanged(bf);
    PSsetcur(bf, bf->curline, bf->curpos);
    ESshowall(bf);
  }
}

/* replaceCurrentWord: ********************************************************
 * replace the current word (the word under or just befor the cursor) with the 
 * given word (newWord). True is returned if the word has been successfully 
 * replaced, False otherwise.
 *****************************************************************************/

Boolean replaceCurrentWord(editor, newWord)
     Editor editor;
     char *newWord;
{
  Boolean locateCurrentWord();
  BufInfo *bf = bufinfoptr(editor);
  char *t = ETtext(bf, bf->curline);
  char *newLine, *lpos, *rpos;
  char c[2];
  int pos1, curline = bf->curline, currow = bf->currow, curcol = bf->curcol;

  if (!t || !newWord)
    return(False);

  if (! locateCurrentWord(editor, &lpos, &rpos))
    return(False);

  c[0] = c[1] = EOS;

  ESshowcur(bf, False, bf->curcol, bf->currow);

  if (rpos < t + strlen(t))
    rpos++;

  c[0] = *rpos;			/* store character */
  if (*rpos)
    *rpos++ = EOS;
  setKillBuffer(bf, strCopy(lpos));

  *lpos = EOS;
  
  newLine = Malloc(strlen(t) + strlen(newWord) + strlen(rpos) + 2);
  
  newLine[0] = EOS;
  strcpy(newLine, t);
  pos1 = strlen(newLine);
  strcat(newLine, newWord);
  bf->curpos = strlen(newLine);
  if (c[0])
    strcat(newLine, c);
  if (*rpos)
    strcat(newLine, rpos);
  
  ETreplace(bf, bf->curline, newLine);
  free(newLine);
  
  ESdisplayLine(bf, bf->curline, pos1);

  setcol(bf);
  if (bf->curline < curline) {
    bf->curline = curline;
    bf->currow = currow;
    bf->curcol = curcol;
    if (ESscroll(bf, topLine(bf, bf->curline) - bf->topline, DIR_UP) < 0)
      setpos(bf); 
  }

  csrChrForget(bf);
  return(True);
}

/* editorSendLine: ************************************************************
 * send a line to the editor and make it believe the user typed the line at the
 * terminal keyboard 
 *****************************************************************************/

editorSendLine(editor, str)
     Editor editor;
     char *str;
{
  register BufInfo *bf = bufinfoptr(editor);
  int fl = bf->flags;

  bf->flags &= ~ED_VIEWMODE;		/* reset view mode temporarily */
  edit(bufinfoptr(editor), str, strlen(str));
  bf->flags = fl;			/* restore flags */
}


/*****************************************************************************
 *              get/set internal editor values and create editor
 *****************************************************************************/

/* editorGotoxy: **************************************************************
 * position the cursor to the given coordinates on the screen;
 *****************************************************************************/

editorGotoxy(editor, x, y)
     Editor editor;
     int x, y;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (x < 0 || x > bf->cols || y < 0 || y > bf->lines)
    return;

  bf->curcol = x;
  bf->currow = y;
  setpos(bf);
}

/* switchCursor: **************************************************************
 * hide or show cursor
 *****************************************************************************/

switchCursor(editor, on)
     Editor editor;
     Boolean on;                 /* if TRUE, show cursor */
{
  register BufInfo *bf = bufinfoptr(editor);

  if (on) 
    bf->flags &= ~ED_CUROFF;
  else
    bf->flags |= ED_CUROFF;
  
/*  ESshowcur(bf, on, bf->curcol, bf->currow);*/
}


/* switchFrame: **************************************************************
 * hide cursor only if mouse is outside of editor window
 *****************************************************************************/

switchFrame(editor, on)
     Editor editor;
     Boolean on;                 /* if TRUE, show frame */
{
  register BufInfo *bf = bufinfoptr(editor);

  if (on) 
    bf->flags &= ~ED_FRAMEOFF;
  else
    bf->flags |= ED_FRAMEOFF;
}

/* needSave: *****************************************************************
 * return True if the editor buffer has been changed since the last save
 ****************************************************************************/

Boolean needSave(editor)
     Editor editor;
{
  return(ETchanged(bufinfoptr(editor)));
}

/* setNeedSave: **************************************************************
 * set the flag that indicates whether the editor buffer needs to be saved
 ****************************************************************************/

void setNeedSave(editor, onoff)
     Editor editor;
     Boolean onoff;
{
  bufinfoptr(editor)->bufPtr->changed = onoff;
}

/* mapEditor: ****************************************************************
 * is called if a window whose descendant is the window containing the editor
 * is either mapped or unmapped. In any case the corresponding flag is set so 
 * that some optimization can be performed (nothing is written to screen 
 * if the editor is unmapped).
 *****************************************************************************/

mapEditor(editor, onoff)
     Editor editor;
     Boolean onoff;
{
  if (onoff)
    bufinfoptr(editor)->flags |= ED_MAPPED;
  else
    bufinfoptr(editor)->flags &= ~ED_MAPPED;
}

/* editorCols: ***************************************************************
 * return the number of cols 
 ****************************************************************************/

editorCols(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->cols);
}

/* editorLines: **************************************************************
 * return the number of lines 
 *****************************************************************************/

editorLines(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->lines);
}

/* editorHeight: *************************************************************
 * return the height in pixels
 *****************************************************************************/

editorHeight(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->height);
}

/* editorWidth: *************************************************************
 * return the width in pixels
 ****************************************************************************/

editorWidth(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->width);
}

/* getEditorFont: ***********************************************************
 * return the font structure
 ****************************************************************************/

XFontStruct *getEditorFont(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->font);
}

/* editorWidget: *************************************************************
 * return the widget of the editor
 ****************************************************************************/

Widget editorWidget(editor)
     Editor editor;
{
  return(bufinfoptr(editor)->w);
}

/* editorWindow: *************************************************************
 * return the window of the editor
 ****************************************************************************/

Window editorWindow(editor)
     Editor editor;
{
  return(XtWindow(bufinfoptr(editor)->w));
}
  
/* exEditSetCursor: *********************************************************
 * set the editor cursor shape and define it if the window has been realised 
 ***************************************************************************/

void exEditSetCursor(editor, cursor)
     Editor editor;
     Cursor cursor;
{
  register BufInfo *bf = bufinfoptr(editor);

  bf->cursor = cursor;
  if (XtWindow(bf->edt))
    XDefineCursor(XtDisplay(bf->w), XtWindow(bf->edt), cursor);
}
  
/* setEditorFile: ************************************************************
 * set the file-name of the specified editor
 ****************************************************************************/

void setEditorFile(editor, file)
     Editor editor;
     char *file;
{
  register BufInfo *bf = bufinfoptr(editor);

  if (bf->filename)
    free(bf->filename);

  bf->filename = strCopy(file);
}

/* switchViewMode: **********************************************************
 * turn view mode on/off
 ****************************************************************************/

switchViewMode(editor, on)
     Editor editor;
     Boolean on;
{
  BufInfo *bf = bufinfoptr(editor);

  if (on)
    bf->flags |= ED_VIEWMODE;
  else
    bf->flags &= ~ED_VIEWMODE;
}

/* editorSetMsg: ************************************************************
 * assigns a message widget to the editor instance
 ***************************************************************************/

editorSetMsg(editor, mw)
     Editor editor;
     Widget mw;
{
  register BufInfo *bf = bufinfoptr(editor);
  bf->msg = mw;
  
  if (mw)
    bf->flags |= ED_MSGWINDOW;
  else
    bf->flags &= ~ED_MSGWINDOW;
}

/* editorMsg: ****************************************************************
 * display a message in the editor`s message window; if no message window 
 * defined, display a notifier window 
 *****************************************************************************/

void editorMsg(editor, str)
     Editor editor;
     char *str;
{
  ESmsg(bufinfoptr(editor), str, True);
}

/* editorClearMsg: ***********************************************************
 * clear the message window of the given editor
 ****************************************************************************/

void editorClearMsg(editor)
     Editor editor;
{
  ESmsgclear(bufinfoptr(editor));			/* clear msg */
}

/* getCmd: *******************************************************************
 * provide the command string to invoke the given function in the parameter
 * cmd; NULL is inserted if the command cannot be located in the command tables
 * associated with the editor.
 ****************************************************************************/

Boolean getCmd(editor, func, cmd)
     Editor editor;
     int (*func)();
     char *cmd;
{
  register BufInfo *bf = bufinfoptr(editor);
  register editCmd *cmdPtr;

  for (cmdPtr = bf->cmdTable; cmdPtr->func; cmdPtr++) {
    if (cmdPtr->func == func) {
      printable(cmdPtr->key, cmd);
      return(True);
    }
  }

  strcpy(cmd, "ESC-");
  cmd += strlen(cmd);

  for (cmdPtr = bf->escTable; cmdPtr->func; cmdPtr++) {
    if (cmdPtr->func == func) {
      printable(cmdPtr->key, cmd);
      return(True);
    }
  }

  strcpy(cmd, "^x");
  cmd += strlen(cmd);

  for (cmdPtr = bf->cxTable; cmdPtr->func; cmdPtr++) {
    if (cmdPtr->func == func) {
      printable(cmdPtr->key, cmd);
      return(True);
    }
  }
  return(False);
}

/* exConfigureEditor: ********************************************************
 * called when a configure notify event has occurred, i.e. the user has 
 * changed the editor window by means of the window manager. The function 
 * calculates the appropriate number of lines and updates the editor structure.
 ****************************************************************************/
/*ARGSUSED*/
void exConfigureEditor(w, event, params, num_params)
     Widget w;
     XConfigureEvent *event;
     String params;
     int *num_params;
{
  configureEditor(w, event, params, num_params);
}

/* makeEditor: ***************************************************************
 * create a new editor instance 
 ****************************************************************************/

Editor makeEditor(w, name, descr, args, cnt)
     Widget w;
     char *name;
     EditorDescr *descr;
     Arg args[];
     int cnt;
{
  return(createEditor(w, name, descr, args, cnt));
}
