/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <X11/Intrinsic.h>
#include <exBasics.h>

#include "exEdit.h"
#include "display.h"

#include <Xm/Xm.h>

#define PS_CHANGED 0x1

static char SccsID[] = "@(#)ps_X.c	1.3\t1/27/93";            /* sccs keyword */

BufInfo *bufinfoptr();

void PSsetcur();
void PSshowcur();
void PSfill();
Boolean isInverse();
void redrawWindow();
void exEditSetScrollbar();
extern upcol();
extern dwncol();
extern botscr();
extern topscr();
extern midscr();
extern upScroll();
extern dwnScroll();
extern Cursor exEditCursor();
extern Cursor exScrollbarCursor();

static GC normal_gc = NULL,
          select_gc = NULL;

/* PSend: *******************************************************************
 * is called when the editor window is destroyed; the buffer
 * structure is cleared 
 ****************************************************************************/
/*ARGSUSED*/
void PSend(w, client_data, call_data)
     Widget w;
     caddr_t client_data;
     caddr_t call_data;
{
  register BufInfo *bf = (BufInfo *) client_data;

  if (! (bf->flags & ED_EXTERNBUFFER))	/* do not clear extern buffer */
    ETclear(bf);

  if (bf->filename)
    free(bf->filename);
  if (bf->name)
    free(bf->name);
  if (bf->bufPtr)
    free(bf->bufPtr);
}

/****************************************************************************
 * PSwrite - write string
 ***************************************************************************/

void PSwrite(bf, row, scol, ncols, str, att)
     register BufInfo *bf;
     int row; 				/* starting row */
     int scol; 				/* starting column */
     int ncols;				/* number of columns to write */
     char *str; 			/* string */
     int att; 				/* attribute */
{
  int len, x, y, col;
  Display *d = XtDisplay(bf->edt);
  Window w = XtWindow(bf->edt);
  GC write_gc;

  if (! (bf->flags & ED_MAPPED))
    return;

  col = scol + bf->lmarg;			/* adjust for left margin */
  if (row < 0 || row >= bf->lines || col >= bf->cols)
    return;

  write_gc = (att == A_NORM) ? normal_gc : select_gc;

  bf->flags |= PS_CHANGED;
  if (col + ncols > bf->cols)
    ncols = bf->cols - col;
  len = strlen(str);
  if (col < 0) {
    if (-col >= len)
      return;
    str = &str[-col];
    ncols += col;
    col = 0;
  }
  ncols = MIN(ncols, len);
  len = ncols * exFontWidth(bf->font);
  x = col * exFontWidth(bf->font);
  y = row * LINEHEIGHT(bf->font) + FONTGAP;
  
  XSetFont(d, write_gc, bf->font->fid);

  XDrawImageString(d, w, write_gc, x, y + bf->baseline, str, ncols);
}

/* PSwrtcells: ****************************************************************
 * write vector of CELLS
 *****************************************************************************

void PSwrtcells(bf, row, col, captr, ncols)
     register BufInfo *bf;
     int row; 			* starting row *
     int col; 			* starting column *
     CELL *captr; 		* CELLs *
     int ncols; 		* number of columns to write *
{
  int i;
  Boolean normal = True;
  int bg, fg;
  register x, y, len;

  if (row < 0 || row >= bf->lines || col < 0 || col >= bf->cols)
    return;
  bf->flags |= PS_CHANGED;
  if (col + ncols > bf->cols)
    ncols = bf->cols - col;
  for (i = 0; i < ncols; i++) {
    if (captr[i].att == A_NORM) {
      if (!normal) {
	fg = exForeground();
	bg = exBackground();
	normal = True;
      }
    }
    else
      if(normal) {	
	bg = exForeground();
	fg = exBackground();
	normal = False;
      }
    
    x = col * exFontWidth(bf->font);
    y = row * LINEHEIGHT(bf->font);
    len = ncols * exFontWidth(bf->font);

    XText(bf->edtwin, x, y + FONTGAP, & captr[i].chr, 1, bf->font->id, fg, bg);
    XPixSet(bf->edtwin, x, y, len, FONTGAP, bg);
  XPixSet(bf->edtwin, x, y + LINEHEIGHT(bf->font) - FONTGAP, len, FONTGAP, bg);
  XPixSet(bf->edtwin, x + exFontWidth(bf->font), y, 1, LINEHEIGHT(bf->font), bg);
  }
}
*/

/* PSclear: ******************************************************************
 * clear to end of row 
 ****************************************************************************/

PSclear(bf, row, col)
     BufInfo *bf;
     int row;
     int col;
{
  RECT rect;

  RASG(&rect, row, col, row, bf->cols);
  PSfill(bf, &rect);
}

/* PSfill: ********************************************************************
 * fill a rectangle
 *****************************************************************************/

void PSfill(bf, srectp)
     register BufInfo *bf;
     RECT *srectp; 			/* rectangle */
{
  RECT srect;
  
  if (! (bf->flags & ED_MAPPED))	/* editor not mapped */
    return;

  if (!intersect(srectp, bf->lines, bf->cols, &srect))
    return;

  bf->flags |= ED_CHANGED;

  XFillRectangle(XtDisplay(bf->edt), XtWindow(bf->edt), select_gc,
	  srect.c1 * exFontWidth(bf->font), 
	  MAX(0, srect.r1 * LINEHEIGHT(bf->font) + 1/* - 2 * FONTGAP */),
	  (srect.c2 - srect.c1) * exFontWidth(bf->font), 
	  (srect.r2 - srect.r1 + 1) * LINEHEIGHT(bf->font));
}

/* PSslide: *******************************************************************
 * slide a rectangle; returns TRUE if slide performed successfully
 *****************************************************************************/

Boolean PSslide(bf, srectp, dir, dist)
     register BufInfo *bf;
     RECT *srectp; 		/* rectangle */
     int dir; 			/* direction */
     int dist; 			/* distance (in screen lines) */
{
  RECT srect;
  int x, y, d;
  unsigned int h, w;
  Display *dsp;
  Window wnd;

  if (! (bf->flags & ED_MAPPED))	/* editor not mapped */
    return(True);
  
  if(!intersect(srectp, bf->lines, bf->cols, &srect) || dist < 1)
    return(True);

  dsp = XtDisplay(bf->w);
  wnd = XtWindow(bf->edt);

  bf->flags |= ED_CHANGED;
  
  x = srect.c1 * exFontWidth(bf->font);
  y = MAX(0, srect.r1 * LINEHEIGHT(bf->font) - 1);
  w = (srect.c2 - srect.c1) * exFontWidth(bf->font) + 2;
  h = (srect.r2 - srect.r1 + 1) * LINEHEIGHT(bf->font);

  switch(dir) {
  case DIR_UP:					/* scroll text area up */
    if (srect.c1 != 0 || srect.c2 != bf->cols)
      return(False);
    d = dist * LINEHEIGHT(bf->font);
    XCopyArea(dsp, wnd, wnd, normal_gc, x, y, w, h, x, y - d);
    XClearArea(dsp, wnd, x, y + h - d, w, d, False);
    return(True);
  case DIR_DOWN:			       	/* scroll text area down */
    if (srect.c1 != 0 || srect.c2 != bf->cols)
      return(False);
    d = dist * LINEHEIGHT(bf->font);
    XCopyArea(dsp, wnd, wnd, normal_gc, x, y, w, h, x, y + d);
    XClearArea(dsp, wnd, x, y, w, d, False);
    return(True);
  case DIR_LEFT:				/* move text area left */
    d = dist * exFontWidth(bf->font);
    XCopyArea(dsp, wnd, wnd, normal_gc, x, y, w, h, x - d, y);
    XClearArea(dsp, wnd, x + w - d, y, d, h, False);
    return(True);
  case DIR_RIGHT:				/* move text area right */
    d = dist * exFontWidth(bf->font);
    XCopyArea(dsp, wnd, wnd, normal_gc, x, y, w, h, x + d, y);
    XClearArea(dsp, wnd, x, y, d, h, False);
    return(True);
  }
  return(False);
}

/* PSstore: *******************************************************************
 * store a rectangle; returns pixmap
 *****************************************************************************/

Pixmap PSstore(bf, srectp, w, h)
     register BufInfo *bf;
     RECT *srectp; 			/* rectangle */
     unsigned int *w, *h;                /* width, height of pixmap */
{
  RECT srect;
  int x, y;
  Display *dsp;
  Window wnd;
  Pixmap pix;

  if (! (bf->flags & ED_MAPPED))	/* editor not mapped */
    return(0);

  if (!intersect(srectp, bf->lines, bf->cols, &srect))
    return(0);

  dsp = XtDisplay(bf->w);
  wnd = XtWindow(bf->edt);

  bf->flags |= ED_CHANGED;
  
  x = srect.c1 * exFontWidth(bf->font);
  y = MAX(0, srect.r1 * LINEHEIGHT(bf->font)) - 1;
  *w = (srect.c2 - srect.c1) * exFontWidth(bf->font) + 2;
  *h = (srect.r2 - srect.r1 + 1) * LINEHEIGHT(bf->font);

  pix = XCreatePixmap(dsp, wnd, *w, *h, DefaultDepthOfScreen(XtScreen(bf->w)));
  XCopyArea(dsp, wnd, pix, normal_gc, x, y, *w, *h, 0, 0);
  XClearArea(dsp, wnd, x, y + *h, *w, *h, False);

  return(pix);
}

/* PSmove: *******************************************************************
 * move a rectangle;
 *****************************************************************************/

PSmove(bf, pix, x, y, w, h)
     register BufInfo *bf;
     Pixmap pix; 		    /* pixmap */
     int x, y;                      /* position of pixmap */
     unsigned int w, h;
{
  Display *dsp;
  Window wnd;

  if (! (bf->flags & ED_MAPPED))	/* editor not mapped */
    return;

/*  if(x und y nicht mehr auf Bildschirm)
    return;
*/

  dsp = XtDisplay(bf->w);
  wnd = XtWindow(bf->edt);

  bf->flags |= ED_CHANGED;

  XCopyArea(dsp, pix, wnd, normal_gc, 0, 0, w, h, x, y);
  XFreePixmap(dsp, pix);
}

/****************************************************************************
 * PSsetcur - set cursor position and adjust scrollbar 
 ***************************************************************************/

void PSsetcur(bf, line, pos)
     BufInfo *bf;
     int line, pos;
{
  char c;
  register Buffer *b = bf->bufPtr;

  if (b->text && b->text[line]) {
    c = b->text[line][pos]; 	/* char under cursor */
    if (c < ' ')
      bf->csrChr = ' ';
    else 
      bf->csrChr = c;
  } else 
    printf("Cannot set cursor (line=%d, pos=%d)(PSsetcur\n", line, pos);
    
  /*exEditSetScrollbar(bf);*/
}

/****************************************************************************
 * PSshowcur - turn cursor on or off
 ***************************************************************************/

void PSshowcur(bf, on, col, row)
     register BufInfo *bf;
     Boolean on; 		/* switch */
     int col, row;
{
  Boolean inverse = isInverse(bf, col, row);

  PSwrite(bf, row, col - bf->lmarg, 1, &(bf->csrChr), 
       on ? (inverse ? A_NORM : A_INVERSE) : (inverse ? A_INVERSE : A_NORM));
}

/* PSshowframe: ***************************************************************
 * converts the solid cursor to a cursor frame 
 *****************************************************************************/

PSshowframe(bf, onoff)
     BufInfo * bf;
     Boolean onoff;
{	
  if (! (bf->flags & ED_MAPPED))		/* editor mapped */
    return;

  if (bf->currow < 0 || bf->currow >= bf->lines || bf->curcol >= bf->cols)
    return;

  XDrawRectangle(XtDisplay(bf->w), XtWindow(bf->edt), 
		 (onoff) ? normal_gc : select_gc, 
         bf->csrFrame.x, bf->csrFrame.y, bf->csrFrame.w, bf->csrFrame.h); 
}

/* isInverse: ****************************************************************
 * returns TRUE if the cursor is in a screen area that is displayed inverse
 ****************************************************************************/

Boolean isInverse(bf, col, row)
     BufInfo *bf;
     int col, row;
{
  int cursorLine, cursor_pos;
  struct Region *r = &bf->region;

  if (! (r->flags & REGION_DISPLAY))
    return(False);

  screen2phys(bf, row, col, &cursorLine, &cursor_pos, False);

  if(r->d_line < 0) 			/* no text selected */
    return(False);
  if(cursorLine < r->d_line)		/* before region */
    return(False);
  if(cursorLine > r->m_line)		/* beyond region */
    return(False);
  if((cursorLine == r->d_line) && (cursor_pos < r->d_pos))
    return(False);
  if((cursorLine == r->m_line) && (cursor_pos > r->m_pos))
    return(False);

  return(True);
}

/* setCmdTable **************************************************************
 * assing the give dwnln command table to the corresponding editor component   
 ****************************************************************************/

static setCmdTable(bf, table, cmdTable)
     BufInfo *bf;
     int table;
     editCmd *cmdTable;
{
  switch (table) {
  case ESC_CMD:
    bf->escTable = cmdTable;
    break;
  case CX_CMD:
    bf->cxTable = cmdTable;
    break;
  case CMD_CMD:
    bf->cmdTable = cmdTable;
    break;
  }
}

/* initCmdTables *************************************************************
 * initialise the editor command tables
 *****************************************************************************/

initCmdTables(editor, cmdTable, escTable, cxTable, addCmds)
     Editor editor;
     editCmd *cmdTable, *escTable, *cxTable;
     cmdDescr *addCmds;
{
  register BufInfo *bf = bufinfoptr(editor);

  setCmdTable(bf, CMD_CMD, cmdTable);
  setCmdTable(bf, ESC_CMD, escTable);
  setCmdTable(bf, CX_CMD, cxTable);

  if (bf->flags & ED_HELP) 		/* reflect new key bindings */
    make_help(bf, addCmds);
}

/* storeFrame: **************************************************************
 * store the corner positions of the frame that is used to draw the 
 * frame-cursor 
 ****************************************************************************/

storeFrame(bf, x0, y0)
     BufInfo *bf;
     int x0, y0;				/* character position */
{
  register w, h, x, y;

  w = exFontWidth(bf->font);
  h = LINEHEIGHT(bf->font);
  x = x0 * w;
  y = MAX(0, y0 * LINEHEIGHT(bf->font) - 1);

  bf->csrFrame.x = x;
  bf->csrFrame.y = y;

  bf->csrFrame.w = w;
  bf->csrFrame.h = h - FONTGAP; 
}

/* redrawEditor: ************************************************************
 * callback function that is called whenever an expose event occurs; the 
 * function redisplays the text in the exposed region of the editor window 
 ****************************************************************************/
/*ARGSUSED*/
void redrawEditor(wd, event, params, num_params)
     Widget wd;
     XExposeEvent *event;
     String params;
     int *num_params;
{
  Arg args[1];
  BufInfo *bf;
  register i, col1, col2, row1, row2;
  int fw, lh, ml, numlines;
  static x = -1, 
         y = -1, 
         w = -1, 
         h = -1;

  if (x < 0)
    x = event->x;
  else
    x = MIN(x, event->x);

  if (y < 0)
    y = event->y;
  else
    y = MIN(y, event->y);

  w = MAX(w, event->width);
  h = MAX(h, event->height);
  
  if (event->count > 0) 	/* collect values from several events */
    return;
  
  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(wd, args, 1);

  XDefineCursor(XtDisplay(bf->w), XtWindow(bf->edt), bf->cursor);
  if (bf->scrollbar)
    XDefineCursor(XtDisplay(bf->w), 
		     XtWindow(bf->scrollbar), exScrollbarCursor());

  bf->flags |= ED_MAPPED;

  fw = exFontWidth(bf->font);
  lh = LINEHEIGHT(bf->font);

  numlines = ETnumlines(bf);
  for (ml = 0, i = bf->topline; i < numlines && ml < bf->lines; i++)
    ml += screenLines(bf, i);
  ml = MIN(bf->lines, ml);
  
  col1 = x / fw;
  col2 = (x + w) / fw;

  row1 = MIN(y / lh, ml);
  row2 = MIN((y + h) / lh, ml);
  
  for (i = row1; i <= row2; i++)		/* display invalid rows */
    ESshowrow(bf, i, col1);
  
  exEditSetScrollbar(bf);			/* adjust the scrollbar */

  x = y = w = h = -1;				/* reset static variables */
}

/* preserveCol: *************************************************************
 * in the case of upcol and dwncol - commands the actual column must be 
 * remembered to restore the col in subsequent col-preserving commands.
 ***************************************************************************/

preserveCol(bf)
     BufInfo *bf;
{
  if (! (bf->oldFnct == upcol  ||
      bf->oldFnct == dwncol    ||
      bf->oldFnct == upScroll  ||
      bf->oldFnct == dwnScroll ||
      bf->oldFnct == botscr    ||
      bf->oldFnct == midscr    ||
      bf->oldFnct == topscr)) {
    bf->preservedCol = bf->curcol;
  }
}

/* setPreservedCol: ***********************************************************
 * set the previously preserved col
 *****************************************************************************/

setPreservedCol(bf)
     BufInfo *bf;
{
  if (bf->oldFnct == upcol ||
      bf->oldFnct == dwncol ||
      bf->oldFnct == upScroll ||
      bf->oldFnct == dwnScroll ||
      bf->oldFnct == botscr ||
      bf->oldFnct == midscr ||
      bf->oldFnct == topscr) {
    bf->curcol = bf->preservedCol;
  }
}

/* ESmsg: *********************************************************************
 * display a message
 *****************************************************************************/

ESmsg(bf, str, forced)
     register BufInfo *bf;
     char *str;
     Boolean forced;              /* if TRUE and no message window exists, 
				     notifier is displayed */
{
  char msg[100];

  if (bf->msg && XtWindow(bf->msg)) {
    XSetFont(XtDisplay(bf->msg), normal_gc, bf->font->fid);
    XClearArea(XtDisplay(bf->msg), XtWindow(bf->msg), 0, 0, 0, 0, False);
    XDrawString(XtDisplay(bf->msg), XtWindow(bf->msg), normal_gc, 
		0, exFontHeight(bf->font), str, strlen(str));
  } else if (forced) {
    sprintf(msg, "editor -- %s --: %s", bf->name, str);
    exAlert(bf->w, -1, -1, msg);
  }
}

void exEditInitGCs(d)
  Display *d;
{
  Window w = XDefaultRootWindow(d);
  normal_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, normal_gc, exWhitePixel(d));
  XSetForeground(d, normal_gc, exBlackPixel(d));
  XSetFunction(d, normal_gc, GXcopy);
  XSetFillStyle(d, normal_gc, FillSolid);
  XSetLineAttributes(d, normal_gc, 1, LineSolid, CapButt, JoinMiter);

  select_gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, select_gc, exBlackPixel(d));
  XSetForeground(d, select_gc, exWhitePixel(d));
  XSetFunction(d, select_gc, GXcopy);
  XSetFillStyle(d, select_gc, FillSolid);
  XSetLineAttributes(d, select_gc, 1, LineSolid, CapButt, JoinMiter);
}

/* statusBarHeight: *********************************************************
 * calciulate the height of the statusbar
 ****************************************************************************/

int msgHeight(editor)
     Editor editor;
{
  BufInfo *bf = bufinfoptr(editor);
  Arg args[3];
  int cnt;
  Dimension bw, st;

  if (! bf->msg)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;

  XtGetValues(bf->msg, args, cnt);

  return(2 * (st + bw) + 3 * exFontHeight(bf->font) / 2);
}

