/*****************************************************************************
 * MDS.h: this file declares the fundamental MenuDataStructure
 ****************************************************************************/

#ifndef _MDS_h
#define _MDS_h

#define SYSMENUS       "esh.menu"	/* file containing system menus */

#define MDS_UMASK 	0777	/* the umask for creating files */

#define CHECK(ptr)      if (! (ptr)) return(setError(ERR_NULL))

#define SYSAPPL         MDS[0]  /* the system application node */

#define MDS_MAXPOPUP   100	/* maximal number of popup menus */
#define MDS_MAXREQ       5	/* maximal number of requests per item */
#define MDS_MAXAPPL	15	/* maximal number of aplication nodes */
#define MDS_MAXMENU     50	/* maximal number of menus per application */

#define MDS_MAXITEM    256      /* maximal number of items per menu */
#define MDS_MAXSTR    1024      /* max. string length to be stored in the MDS */
#define MDS_NAMELEN     80      /* length of a name */

#define MDS_FILESUFFIX  ".menu" /* suffix for file to hold MDS-description */

/* the valid access flags for a menu description are define in exMenu.h */

/* define the valid flags for a menu description */
#define MDS_NOACC       0x0     /* no access yet specified */
#define MDS_READACC	0x1	/* allowed ro read the MDS */
#define MDS_CHNGACC	0x2	/* allowed to make temporar changes */
#define MDS_TEMP        0x4     /* temporary menu, will not be saved */
#define MDS_SYS         0x8     /* system menu */

/* excerped from sys/types.h to handle more than 32 activity bits */

#define	NBBY	8		/* number of bits in a byte */
#define	ACTMASK_SIZE	MDS_MAXITEM		/* one bit for each item */
#define	ACTMASK_BITS	(sizeof(long) * NBBY)	/* bits per mask */

typedef	struct act_mask {
	long actmask_bits[(ACTMASK_SIZE + ACTMASK_BITS - 1) / ACTMASK_BITS];
} act_mask;

#define	AM_SET(p, n)	((p)->actmask_bits[(n)/ACTMASK_BITS] \
			 	|= (1 << ((n) % ACTMASK_BITS)))
#define	AM_CLR(p, n)	((p)->actmask_bits[(n)/ACTMASK_BITS] \
			 	&= ~(1 << ((n) % ACTMASK_BITS)))
#define	AM_ISSET(p, n)	((p)->actmask_bits[(n)/ACTMASK_BITS] \
			 	& (1 << ((n) % ACTMASK_BITS)))
#define	AM_ZERO(p)	bzero((char *)(p), sizeof(*(p)))

#define AM_ONES(p)	memset((char *) (p), -1, sizeof(*(p)))

#define AM_CPY(s, d)	bcopy((char *) (s), (char *) (d), sizeof(*(s)))

/* item specific flags */

#define MH_NODEFAULT         0x0
#define MH_DEFAULTITEM	     0x1	/* item is initially selected */
#define MH_TOGGLEITEM        0x2	/* item is a toggle item */
#define MH_TICKON            0x4	/* tick is displayed */
/* constants that are used to describe item requests */

#define MH_MDSFUNCTION 	     0x0	/* mds defined function */
#define MH_USERFUNCTION      0x1	/* user defined function */

/* constants that are used during editing of menu descriptions (see: menu.c) */
#define M_CREATE   0
#define M_MODIFY   1

/* define some string constants */
#define MDS_M_HELP 	"no help available"
#define MDS_I_HELP 	"no help available"

typedef char *id;

/* define the menu types */
typedef enum { 			
        MDS_M_NULL    = 100,
	MDS_M_POPUP   = 101,	/* pop-up menu */
	MDS_L_POPUP   = 102,	/* pop-up list */
	MDS_M_PULLD   = 103,	/* pull-down menu */
	MDS_L_PULLD   = 104,	/* pull-down list */
} MDS_MENU_TYPE;

/* define the item selection types to determine, how the item selection count 
 * is to be interpreted 
 */

/* define the item selection types to determine, how the item selection count 
 * is to be interpreted */
typedef enum {
        MDS_CNT_NULL = 200,
	MDS_CNT_EQ   = 201,	/* no. of items to be selected must be equal*/
	MDS_CNT_LT   = 202,	/* less */
	MDS_CNT_LE   = 203,	/* less or equal */
	MDS_CNT_GT   = 204,	/* greater */
	MDS_CNT_GE   = 205,	/* greater or equal than the selection count */
} MDS_CNT_TYPE;

/* define the different types of representing a menu item */
typedef enum {
        MDS_RTYP_NULL   = 300,
	MDS_RTYP_STRING = 301,	/* item is represented as string */
} MDS_REPR_TYPE;

/* define the legal colors that may be used for defining menus */
typedef enum {
   MDS_NOCOLOR  	= 400,
   MDS_WHITE    	= 401,
   MDS_BLACK    	= 402,
   MDS_GREY     	= 403,
   MDS_LIGHTGREY	= 404,
   MDS_DARKGREY 	= 405,
   MDS_ULTRALIGHTGREY 	= 406,
} MDS_COLOR;

/* define how a position is interpreted */
typedef enum {
  MDS_POS_NULL  = 500,
  MDS_POS_ABS   = 501, 		/* absolute position */
  MDS_POS_CREL  = 502,		/* relative to cursor position */
  MDS_POS_RECT  = 503,		/* x/y are interpreted as width and height */
} MDS_POS_TYPE;

typedef struct _Pos {
  int x, y;
  MDS_POS_TYPE posType;
} Pos;

/* the structure MHfunc defines the functions that an application provides 
   to the menu - handling system, so that they can be called directly by 
   menu-selection. The structure consists of a pointer to the function to be 
   called and a string that holds an identification to be used during 
   menu edititing;
*/

typedef struct _MHfunc {
  int (*func)();		/* pointer to the function to be called */
  char * funcName;		/* user representation for menu editor */
} MHfunc;

/* RequestParm defines the parameter interface for functions that are called 
 * by the menu handling system to satisfy a request 
 */
typedef struct _RequestParm {
  caddr_t menu;			/* the associated menu */
  Widget parent;	       	/* the parent widget */
  char *data;			/* additional data */
  caddr_t ptr;			/* points to add. data */
  int x, y;			/* pos */
  char *buffer;			/* buffer for return data */
  int size;			/* size */
  int cnt;			/* number of tokens requested */
  MDS_CNT_TYPE cntType;		/* count Type */
} RequestParm;

#define MDS_MAXFUNC	100 	/* size of the MDS function table */

/* constants that define the index of the corresponding function in the
   MDS_functions array */

#define DSPLINWND  1
#define REQVIAWND  2
#define REQVIAMENU 3
#define REQVIAFORM 4

#define MDS_FUNCT_CNT 4

/*****************************************************************************
 * each applicationNode holds the application relevant information; it 
 * establishes references to the application name and a handle of all menus
 ****************************************************************************/

typedef struct _ApplicationNode {
  Boolean _changed;		/* is true, if one of the components has been 
				   modified */
  char *applicationName;	/* name of the application */
  int a_access;			/* access rights for appliocationNode */
  int menuCnt;			/* number of menus for this application */
  struct _MHmenu **m_table;	/* the menu table */
  char *file;			/* file where application node is stored */
				/* array of pointers to function structures */
  MHfunc ApplFunc[MDS_MAXFUNC]; 
} ApplicationNode;

/* this is the basi menu-structure for each menu; it holds menu-relevant 
   information and a pointer to the list of menu-items
*/
typedef struct _MHmenu {
  struct _MHmRepr *m_repr;	/* menu representation structure */
  MDS_MENU_TYPE m_type;		/* menu-type (e.g. popup, pulldown, ..) */
  int m_flags;			/* menu flags: access rights, state of menu 
				   description (eg. temporary, permanent) */
  int itemCnt;			/* number of items for this menu */
  int maxItem;			/* max. number of to be displayed */
  int i_selCnt;			/* how many items can be selected */
  MDS_CNT_TYPE i_selType;	/* specification of selCnt; e.g lt, gt ...) */
  act_mask activityMask;	/* activity mask */
  struct _Pos m_pos;		/* menu position */
  char m_key;			/* char to select the menu */
  struct _MHitem **i_table;	/* item table */
  struct _MHhlp *m_hlp;		/* what to do if help is requested */
} MHmenu;

/* the help structure */

typedef struct _MHhlp {
  char *hlpmsg;		  /* the message to be displayed */
  int (* hlpFnct)();	  /* the function to display the message */
} MHhlp;

/* m_repr defines how a menu will be represented on the screen */

typedef struct _MHmRepr {
  MDS_REPR_TYPE reprType;  /* type for representation (e.g. text, icon, ..) */
  char *name; 		   /* a pointer to the representation data */
} MHmRepr;

/* the item data-structure */
typedef struct _MHitem {
  int flag;                     /* flags for kind of item */
  struct _MHiRepr *i_repr;	/* item representation */
  struct _Pos i_Pos;		/* item position */
  act_mask activityMask;	/* activity mask */
  struct _MHhlp *i_hlp;		/* if help is requested by the user */
  struct _MHiRet *i_return;	/* the structure that contains the value to be
				 * returned upon selection of this item */
  struct _MHiReq *i_requests[MDS_MAXREQ];
				/* a structure for each request to get 
				 * more information upon item selection */
} MHitem;

/* the structure that defines the representatio of an item normal (not selected
   but selectable), selected and not selectable 
*/
typedef struct _MHiRepr {
  char *name;		     	/* item name */
  char key;			/* the item key */
} MHiRepr;

/* the structure that defines the return value provided by an item upon 
   selection 
*/
typedef struct _MHiRet {
  char *retPattern;		/* pattern of the message to be returned */
  int reqCnt;		 	/* number of requests */
} MHiRet;

/* the datatype for requesting additional information upon selection */
typedef struct _MHiReq {
  int flags;		   /* user defined function, etc... */
  int size;		   /* size (in case of reqViaWindow the length of the 
			    * entry */
  int (* reqFnct)();	   /* function to request a value from the user */
  char *initialValue;	   /* initial value for this request */
  char *prelimInitialValue;/* stores new initial values after item selection;
			    * is copied to initialValues after successful menu
			    * completion */
  caddr_t reqPtr;	   /* points to data for the request function */
  char *reqData; 	   /* data for the requesting function: e.g. in the 
			    * case of reqViaWindow a message,  in the case 
			    * of a menu or form the menu's or form's name */
  int cnt;		   /* number of tokens to be requested */
  MDS_CNT_TYPE reqCntType; /* defines how the cnt is to be interpreted 
			    * (e.g. <, >, <=, ...) */
  char *returnValue;	   /* holds the return value in case of a regular 
			      request-function; this information is used to 
			      insert return-values after all user-functions 
			      have been executed (see requestFromUser and 
			      assembleReturnValues in exMenu.c
			    */
} MHiReq;


typedef struct _applId {
  char *name;           /* application name */
  char *file;           /* file in which application node is stored */
} applId;


/****************************************************************************
 *                  external declarations 
 ***************************************************************************/

extern MHfunc MH_FUNCTIONS[];	/* the table of user provided function */
extern MHfunc MDS_functions[];  /* will conatain the predefined functions */
extern ApplicationNode *MDS[];  /* array of current applications */

#endif _MDS_h
