/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/*****************************************************************************
 * This file contains the application functions of the macro editor listed in
 * menu "macro"
 *****************************************************************************/

#include <stdio.h>
#include <strings.h>
#include <exBasics.h>
#include <exEdit.h>
#include <exStxt.h>
#include <exLayout.h>
#include <exMenu.h>
#include <exMacros.h>
#include <exEntry.h>
#include <ife.h>
#include "macroEdit.h"
#include <exErrors.h>

static char SccsID[] = "@(#)macfuncs.c	1.2\t11/13/92";            /* sccs keyword */

extern macroEditor *MacroEditor;

static char msg[200];
static char fullName[256];

static loadSetInEditor();

/*****************************************************************************
 *
 * 		     		OPTIONS
 *
 ****************************************************************************/

/* quitMacroEditor: ***********************************************************
 * quit the macro editor 
 *****************************************************************************/

quitMacroEditor(parms)
     RequestParm *parms;
{
  if (MacroEditor->flags & ME_CONNECT) {
    exSendMsg(fileno(stdout), getpid(), IFE_INFO, IFE_STOP);
    sleep(2);
  }
  XCloseDisplay(XtDisplay(parms->parent));
  exit(1);
}

/*****************************************************************************
 *
 * 		     		MACRO SET MENU
 *
 ****************************************************************************/

/* newMacroSet: ***************************************************************
 * create a new macro set and make it the current one 
 *****************************************************************************/

newMacroSet(parms)
     RequestParm *parms;
{
  char *macroSet, *dirName;
  char *makeMacroDirPath();

  if (MacroEditor->macroSet && needSave(MacroEditor->editor)) {
    sprintf(msg, "macro -- %s --\nof macro set -- %s --\nhas not been saved yet!\nSave changes?", MacroEditor->macroName, MacroEditor->macroSet);

    switch (exNotify2User(parms->parent, -1, -1, msg)) {
    case ANSWERYES:
      saveMacro(parms);
      break;
    case ANSWERCANCEL:
      return(-1);
    default:
      break;
    }
  }

  macroSet = 
    exGetEntry(parms->parent, -1, -1, "Enter macro set name: ", NULL, 20);
 
  if (! macroSet || !*macroSet)
    return(-1);

  dirName = makeMacroDirPath(macroSet);

  if (macroSetExists(dirName)){
    sprintf(msg, "Warning: macro set -- %s -- already exists!\nSelect another name want", macroSet);
    exAlert(parms->parent, -1, -1, msg);
    return(-1);
  }
  else 
    sprintf(msg, "New current macro set: -- %s -- !", macroSet);

  if (! fileExists(macroSet)) {		/* subdirectory does not exist */
    if (mkdir(macroSet, 0755) < 0) {	/* create it */
      setError(ERR_CREATE);
      return(-1);
    }
  }

  macroMsg(msg);
  clearMacroBuffer();
    
  free(MacroEditor->macroSet);
  free(MacroEditor->macroName);

  MacroEditor->macroName = (char *) 0;
  MacroEditor->macroSet = dirName;

  free(macroSet);

  setStatus(dirName, NULL);

  return(0);
}

/* openMacroSet: *************************************************************
 * show the user's macro-sets for selection 
 ****************************************************************************/

openMacroSet(parms)
     RequestParm *parms;
{
  char *macroSet;

  if (MacroEditor->macroSet && needSave(MacroEditor->editor)) {
    sprintf(msg, "macro -- %s --\nof macro set -- %s --\nhas not been saved yet!\nSave changes?", MacroEditor->macroName, MacroEditor->macroSet);

    switch (exNotify2User(parms->parent, -1, -1, msg)) {
    case ANSWERYES:
      saveMacro(parms);
      break;
    case ANSWERCANCEL:
      return(-1);
    default:
      break;
    }
  }

  if (! (macroSet = selectMacroSet(parms->parent)))
    return(-1);

  loadSetInEditor(macroSet);

  return(0);
}

/* closeMacroSet: *************************************************************
 * save the current macroSet
 ****************************************************************************/
  
closeMacroSet(parms)
     RequestParm *parms;
{

  if (! MacroEditor->macroSet || ! *MacroEditor->macroSet) {
    macroMsg("No current macro set open !");
    return(-1);
  }

  if (needSave(MacroEditor->editor))
    saveMacro(parms);

  return(0);
}

/* saveAsMacroSet: ***********************************************************
 * duplicate the current macroSet
 ****************************************************************************/
  
saveAsMacroSet(parms)
     RequestParm *parms;
{
  char *macroSet, *dirName, *buffer, *txt;
  char *makeMacroDirPath();
  int nb;

  if (! MacroEditor->macroSet || ! *MacroEditor->macroSet) {
    macroMsg("No current macro set open !");
    return(-1);
  }

  macroSet = 
    exGetEntry(parms->parent, -1, -1, "Enter macro set name: ", NULL, 20);

  if (! macroSet || !*macroSet)
    return(-1);

  dirName = makeMacroDirPath(macroSet);

  if (macroSetExists(dirName)) {
    sprintf(msg, "Warning:\nmacro set -- %s -- already exists!\nDo you want to overwrite existing set?", macroSet);
    if (exNotify2User(parms->parent, -1, -1, msg) != ANSWERYES)
      return(-1);
  }
  else {
    if (mkdir(dirName, 0755) < 0) {	/* create new macro set */
      setError(ERR_CREATE);
      return(-1);
    }
    sprintf(msg, "New current macro set: -- %s -- !", macroSet);
  }

  macroMsg(msg);

  /* generate shell command to copy correspoding files using system(3) */
  sprintf(msg, "cp %s/*.mac %s", MacroEditor->macroSet, dirName);
  system(msg);
    
  txt = getBufferContents(MacroEditor->editor);

  txt[MAX(strlen(txt) - 1, 0)] = EOS;		/* discard last <cr> */

  if (needSave(MacroEditor->editor)) {
    storeMacro(MacroEditor->me_shell, dirName, MacroEditor->macroName, txt);
    sprintf(msg, "macro -- %s -- has not been saved yet ! Save changes ?",
	    MacroEditor->macroName);

    switch (exNotify2User(parms->parent, -1, -1, msg)) {
    case ANSWERYES:
      saveMacro(parms);
      break;
    case ANSWERCANCEL:
      return(-1);
    default:
      break;
    }
  }

  sprintf(msg, "macro set -- %s -- closed !", MacroEditor->macroSet); 
  macroMsg(msg);

  free(MacroEditor->macroSet);
  MacroEditor->macroSet = dirName;

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);

  /* load macro */
  if (makeFullName(MacroEditor->macroSet, MacroEditor->macroName, fullName, 
		    sizeof(fullName)) < 0)
    return(-1);

  nb = loadMacroText(fullName, &buffer);     
  openMacroEditor((Widget) 0, (caddr_t) 0, (caddr_t) 0);
  if (nb < 0)
    sprintf(msg, "Macro -- %s -- not loaded", MacroEditor->macroName);
  else {
    fillMacroBuffer(MacroEditor->macroName, buffer);
    sprintf(msg, "Macro -- %s -- loaded (%d bytes)", MacroEditor->macroName,
	    nb);
  }
  
  MacroEditor->flags &= ~MACRO_PROC;
  macroMsg(msg);
  free(buffer);

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);

  touchFile(MacroEditor->macroSet, time(0), time(0));

  return(0);
}

/* deleteMacroSet: **********************************************************
 * delete an entire macro set; rebuild the corresponding menus;
 ***************************************************************************/

deleteMacroSet(parms)
     RequestParm *parms;
{
  register char *macroSet;
  int cnt;
  char *ms;

  macroSet = selectMacroSet(parms->parent);

  if (! macroSet || ! *macroSet)
    return(-1);

  sprintf(msg, "Do you really want to\ndelete macro set <%s>?", 
	  MacroEditor->macroSet);
  if (exNotify2User(parms->parent, -1, -1, msg) != ANSWERYES)
    return(-1);

  cnt = removeMacroSet(macroSet); 
  
  if (cnt < 0) {
    sprintf(msg, "Macro set (%d macros) -- %s -- not removed !",-cnt,macroSet);
    return(-1);
  }
  else 
    sprintf(msg, "Macro set (%d macros) -- %s -- removed !", cnt, macroSet);

  macroMsg(msg);

  if (strcmp(macroSet, MacroEditor->macroSet) == 0) {   /* current macro set 
							 removed */
    if (ms = getNextMacroSet(macroSet)) 
      loadSetInEditor(ms);
    else
      loadInitialSet();
  }

  return(0);
}

/* displayMacroSets: **********************************************************
 * display a list of the user's macro sets
 *****************************************************************************/

displayMacroSets(parms)
     RequestParm *parms;
{
  selectMacroSet(parms->parent);
  setStatus(NULL, NULL);
  return(0);
}

/******************************************************************************
 *
 * 		     		MACRO MENU
 *
 *****************************************************************************/


/* newMacro: *****************************************************************
 * prepare the macro editor for editing a new macro definition 
 *****************************************************************************/

/*ARGSUSED*/
newMacro(parms)
     RequestParm *parms;
{
  char *macro;
  int answer = 0;

  if (! MacroEditor->macroSet) {
    macroMsg("No current macroset: open a set first !");
    return(-1);
  }

  if (MacroEditor->macroSet && needSave(MacroEditor->editor)) {
    sprintf(msg, "macro -- %s --\nof macro set -- %s --\nhas not been saved yet!\nSave changes?", MacroEditor->macroName, MacroEditor->macroSet);

    if ((answer = exNotify2User(parms->parent, -1, -1, msg)) == ANSWERYES)
      saveMacro(parms);
    else if (answer == ANSWERCANCEL)
      return(-1);
  }

  macro = exGetEntry(parms->parent, -1, -1, "Enter the macro name: ", NULL, 
		     20);
    
  if (! macro || !*macro)
    return(-1);

  free(MacroEditor->macroName);
  MacroEditor->macroName = macro;

  MacroEditor->flags &= ~MACRO_PROC;

  clearMacroBuffer();

  openMacroEditor((Widget) 0, (caddr_t) 0, (caddr_t) 0);
  macroMsg("Ready to enter a macro definition !");

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);

  return(0);
}

/* loadMacro: *****************************************************************
 * load the specified macro definition from the macro file and store it into 
 * the buffer of the macro editor.
 *****************************************************************************/

loadMacro(parms)
     RequestParm *parms;
{
  char *macroName;
  char *buffer, *name;
  int nb, answer = 0;

  if (MacroEditor->macroSet && needSave(MacroEditor->editor)) {
    sprintf(msg, "macro -- %s --\nof macro set -- %s --\nhas not been saved yet!\nSave changes?", MacroEditor->macroName, MacroEditor->macroSet);

    if ((answer = exNotify2User(parms->parent, -1, -1, msg)) == ANSWERYES)
      saveMacro(parms);
    else if (answer == ANSWERCANCEL)
      return(-1);
  }

  macroName = selectMacro(parms->parent, MacroEditor->macroSet, False);
  if (! macroName) {
    macroMsg("Cannot locate your macros !");
    return(-1);
  }
  
  nb = loadMacroText(macroName, &buffer);      	/* load macro into buffer */

  /* extract macro name from full-qualified name */
  if ((name = rindex(macroName, '/')) == (char *) 0)
    name = "??";
  else 
    *name++ = EOS;

  openMacroEditor((Widget) 0, (caddr_t) 0, (caddr_t) 0);
  if (nb < 0)
    sprintf(msg, "Macro -- %s -- not loaded", name);
  else {
    fillMacroBuffer(macroName, buffer);
    sprintf(msg, "Macro -- %s -- loaded (%d bytes)", name, nb);
  }
  
  MacroEditor->flags &= ~MACRO_PROC;
  free(MacroEditor->macroName);
  MacroEditor->macroName = strCopy(name);
  macroMsg(msg);
  free(buffer);
  free(macroName);

  setStatus(MacroEditor->macroSet, name);

  return(0);
}

/* saveMacro: *****************************************************************
 * save the contents of the macro editor buffer to where it has been read from.
 * If a new macro has been defined, ask the user where to store 
 *****************************************************************************/

saveMacro(parms)
     RequestParm *parms;
{
  char *macro, *txt, *macroSet, *dirName;

  if (! MacroEditor->macroSet) {
    macroMsg("No current macroset: open a set first !");
    return(-1);
  }
  else if (strcmp(MacroEditor->macroSet, INITIALSET) == 0) {
    macroSet = 
      exGetEntry(parms->parent, -1, -1, "Enter macro set name: ", NULL, 20);
 
    if (! macroSet || !*macroSet)
      return(-1);

    dirName = makeMacroDirPath(macroSet);

    if (macroSetExists(dirName)) 
      sprintf(msg, "Warning: macro set -- %s -- already exists !", macroSet);
    else 
      sprintf(msg, "New current macro set: -- %s -- !", macroSet);

    macroMsg(msg);
    
    free(MacroEditor->macroSet);
    MacroEditor->macroSet = dirName;
    setStatus(dirName, MacroEditor->macroName);

    free(macroSet);
  }

  if  (! MacroEditor->macroName) {
    macro = 
      exGetEntry(parms->parent, -1, -1, "Enter the macro name: ", NULL, 20);
    if (! macro || !*macro)
      return(-1);
    
    free(MacroEditor->macroName);
    MacroEditor->macroName = macro;
  }
  
  txt = getBufferContents(MacroEditor->editor);

  txt[MAX(strlen(txt) - 1, 0)] = EOS;		/* discard last <cr> */

  if (storeMacro(MacroEditor->me_shell, MacroEditor->macroSet, 
	     				MacroEditor->macroName, txt) < 0 ) 
    sprintf(msg, "Macro -- %s -- NOT stored !", MacroEditor->macroName);
  else
    sprintf(msg, "Macro -- %s -- stored (%d bytes) !", 
	    MacroEditor->macroName, strlen(txt));
  macroMsg(msg);

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);

  return(0);
}

/* saveMacroAs: ************************************************************
 * duplicate the contents of the macro editor buffer and store it as a macro 
 * under a name requested from the user. 
 *****************************************************************************/

saveMacroAs(parms)
     RequestParm *parms;
{
  char *macro, *txt;

  if (! MacroEditor->macroSet) {
    macroMsg("No current macroset: open a set first !");
    return(-1);
  }

  macro = exGetEntry(parms->parent, -1, -1, "Enter the macro name: ", NULL, 
		     20);
    
  if (! macro || !*macro)
    return(-1);
    
  free(MacroEditor->macroName);
  MacroEditor->macroName = macro;
  
  txt = getBufferContents(MacroEditor->editor);

  if (storeMacro(MacroEditor->me_shell, MacroEditor->macroSet, 
		 MacroEditor->macroName,txt)< 0)
    sprintf(msg, "Macro -- %s -- NOT stored !", MacroEditor->macroName);
  else
    sprintf(msg, "Macro -- %s -- stored (%d bytes)", MacroEditor->macroName,
	    strlen(txt));
  macroMsg(msg);

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);

  return(0);
}

/* saveMacroInSet: ************************************************************
 * save the contents of the macro editor buffer to a macro set that is 
 * requested from the user. 
 *****************************************************************************/

saveMacroInSet(parms)
     RequestParm *parms;
{
  char *macroSet, *macro, *txt;

  macroSet = selectMacroSet(parms->parent);
  if (! macroSet || ! *macroSet)
    return(-1);

  if  (! MacroEditor->macroName) {
    macro = 
      exGetEntry(parms->parent, -1, -1, "Enter the macro name: ", NULL, 20);
    if (! macro || !*macro)
      return(-1);
    
    free(MacroEditor->macroName);
    MacroEditor->macroName = macro;
  }
  
  txt = getBufferContents(MacroEditor->editor);

  if (storeMacro(MacroEditor->me_shell,macroSet,MacroEditor->macroName,txt)< 0)
    sprintf(msg, "Macro -- %s -- NOT stored !", MacroEditor->macroName);
  else
    sprintf(msg, "Macro -- %s -- stored (%d bytes)", MacroEditor->macroName,
	    strlen(txt));
  macroMsg(msg);

  setStatus(macroSet, macro);

  return(0);
}

/* deleteMacros: *************************************************************
 * delete macros that have been selected by the user
 ****************************************************************************/

deleteMacros(parms)
     RequestParm *parms;
{
  register char *macroFile, *macroSet, *macro, *delmacro, *ptr;
  int rc, i = 0;

  macroFile = selectMacro(parms->parent, MacroEditor->macroSet, True);
  if (! macroFile)
    return(-1);

  sprintf(msg, "Do you really want to\ndelete all selected macros?");
  if (exNotify2User(parms->parent, -1, -1, msg) != ANSWERYES)
    return(-1);

  ptr = macroFile;
  while (ptr && i < MDS_MAXMENU) {    
    if (strncmp(ptr, MacroEditor->macroSet, strlen(MacroEditor->macroSet))
	== 0) {
      ptr += strlen(MacroEditor->macroSet) + 2;
      if (ptr = (char *)index(ptr, '/'))
	delmacro = makeStr(macroFile, ptr,strlen(macroFile) - strlen(ptr) + 1);
      else
	delmacro = macroFile;

      rc = removeMacroFile(parms->parent, delmacro); 
  
      if ((macro = rindex(delmacro, '/')) == (char *) 0)  
	macro = "??";
      else 
	*macro++ = EOS;                               /* get macro name */
  
      if ((macroSet = rindex(macroFile, '/')) == (char *) 0) 
	macroSet = "??";
      else 
	macroSet++;                                   /* get macro set name */
  
      if (rc < 0) 
	sprintf(msg, "Macro -- %s -- not removed !", macro);
      else if (rc == 1)
	sprintf(msg, "Macro -- %s -- removed from macroSet -- %s -- !", 
		macro, macroSet);
      else if (rc == 2)
	sprintf(msg, "Last macro -- %s --; menu -- %s -- removed !", 
		macro, macroSet);
      else 
	sprintf(msg, "Internal problems: macro -- %s -- probably not deleted !", 
		macro);

      macroMsg(msg);

      macroFile = ptr;
    }
    i++;
  }

  return(0);
}

/* displayMacros: ************************************************************
 * display the macros in the current set
 *****************************************************************************/

displayMacros(parms)
     RequestParm *parms;
{
  if (! MacroEditor->macroSet || ! *MacroEditor->macroSet) {
    macroMsg("No current macro set !");
    return(-1);
  }

  selectMacro(parms->parent, MacroEditor->macroSet, False);

  return(0);
}

/******************************************************************************
 *
 * 		     	    UTILITY FUNCTIONS
 *
 *****************************************************************************/

/* loadSetInEditor: **********************************************************
 * load a macro set into the macro editor; first macro of this set is shown
 ****************************************************************************/

static loadSetInEditor(macroSet)
     char *macroSet;
{
  char *buffer;
  int nb;

  if (! macroSet)
    return;
  
  free(MacroEditor->macroSet);
  free(MacroEditor->macroName);
  MacroEditor->macroName = (char *) 0;
  MacroEditor->macroSet = macroSet;
  setStatus(macroSet, NULL);

  MacroEditor->macroName = findFirstMacro(macroSet);

  if (!MacroEditor->macroName) {
    setStatus(MacroEditor->macroSet, NULL);
    return;
  }
    
  /* load macro */
  if (makeFullName(macroSet, MacroEditor->macroName, fullName, 
		    sizeof(fullName)) < 0)
    return;

  nb = loadMacroText(fullName, &buffer);     
  openMacroEditor((Widget) 0, (caddr_t) 0, (caddr_t) 0);
  if (nb < 0)
    sprintf(msg, "Macro -- %s -- not loaded", MacroEditor->macroName);
  else {
    fillMacroBuffer(MacroEditor->macroName, buffer);
    sprintf(msg, "Macro -- %s -- loaded (%d bytes)", MacroEditor->macroName,
	    nb);
  }
  
  MacroEditor->flags &= ~MACRO_PROC;
  macroMsg(msg);
  free(buffer);

  setStatus(MacroEditor->macroSet, MacroEditor->macroName);
}

/* loadInitialSet: ************************************************************
 * create and load initial macro set
 *****************************************************************************/

loadInitialSet()
{
  sprintf(msg, "Initial macro set: -- %s -- !", INITIALSET);

  macroMsg(msg);

  free(MacroEditor->macroSet);
  free(MacroEditor->macroName);

  MacroEditor->macroName = NULL;
  MacroEditor->macroSet = strCopy(INITIALSET);

  setStatus(MacroEditor->macroSet, NULL);
}
